/*
 * Copyright (c) 2004-2005  
 *              Martin Reinecke (1), Klaus Dolag (1)
 *               (1) Max-Plank-Institute for Astrophysics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include<iostream>
#include<cmath>
#include<fstream>
#include<algorithm>

#include "arr.h"
#include "cxxutils.h"
#include "paramfile.h"
#include "kernel/bstream.h"
#include "kernel/colour.h"
#include "config/config.h"
#include "utils/colourmap.h"

using namespace std;
using namespace RAYPP;

const float8 Pi=3.14159265358979323846;

struct PARTICLE
  {
  float4 x,y,z,r,ro,T;
  int type;
  };

struct zcmp
  {
  int operator()(const PARTICLE &p1, const PARTICLE &p2)
    {
    return p1.z>p2.z;
    }
  };

class exptable
  {
  private:
    int nexp;
    float8 maxexptab;
    float8 expfac;
    float8 *exptab;

  public:
    exptable (int nexp_, float8 maxexp_)
      : nexp(nexp_), maxexptab(maxexp_), expfac(nexp/maxexptab),
        exptab(new float8[nexp])
      {
      for (int m=0; m<nexp; ++m)
        exptab[m]=exp(m/expfac);
      }

    float8 operator() (float8 arg) const
      {
      arg *= expfac;
      int iarg=int(arg);
      if (iarg<0) return 1.;
      if (iarg>=nexp-1) return 0.;
      float8 frac=arg-iarg;
      return (1.-frac)*exptab[iarg]+frac*exptab[iarg+1];
      }
  };

int gadget_find_block(bifstream &file,const string &label)
  {
  int4 blocksize=0, blksize;
  char blocklabel[5]={"    "};

  file.rewind();

  while(!file.eof() && blocksize == 0)
    {
    file >> blksize;
    if (blksize != 8)
      {
      cout << "wrong structure in GADGET file" << endl;
      exit(1);
      }
    else
      {
      file.get(blocklabel,4);
      blocklabel[4] = 0;
//cout << "blocklabel: " << blocklabel << endl;
      file >> blocksize;
//cout << "blocksize: " << blocksize << endl;
      file.skip(4);
      if (label!=blocklabel)
	{
        file.skip(blocksize);
        blocksize=0;
        }
      }
    }
  return(blocksize-8);
  }

int gadget_read_header(bifstream &file, int *npart,double *massarr,
  double &time,double &redshift, int *npartall)
  {
  int blocksize = gadget_find_block (file,"HEAD");
  planck_assert (blocksize>0, "Header block not found");
  file.skip(4);
  file.get(npart,6);
  file.get(massarr,6);
  file >> time >> redshift;
  file.skip(8);
  file.get(npartall,6);

  return blocksize;
  }

int main (int argc, const char ** argv)
  {
const float8 Pi=3.14159265358979323846;
  announce ("splotch");
  planck_assert (argc==2,"usage: splotch <parameter file>");
  paramfile params (argv[1]);

  cout << "initializing" << endl;
  int res = params.find<int>("resolution",200);
  string infilename = params.find<string>("infile");
  string outfile = params.find<string>("outfile");
  string filename;
  int numfiles = params.find<int>("numfiles",1);
  int lable_nr = params.find<int>("lable_nr",1);
  string lable;
  switch(lable_nr)
    {
    case 1:lable="TEMP"; break;
    case 2:lable="RHO "; break;
    }
  int colors = params.find<int>("colors",1);
  bool log10 = params.find<bool>("log",true);
  bool stars = params.find<bool>("stars",false);
  float star_hsml = params.find<float>("star_hsml",5.0);
  float minval = params.find<float>("min",1024);
  float maxval = params.find<float>("max",1024);
  
  float8 brightness = params.find<double>("brightness",1.);

  exptable xexp(10000,-20.);

  int4 npart[6],npartall[6];
  float8 massarr[6],time,redshift;

  if(numfiles>1) filename=infilename+"."+dataToString(0);
  else           filename=infilename;
  bifstream infile(filename.c_str(),file_is_msb);
  if (!infile) {cout << "could not open input file! <" << filename << ">" << endl; exit(1);}
  gadget_read_header(infile,npart,massarr,time,redshift,npartall);
  if(!stars) npartall[4]=0;
  cout <<  time << " " << redshift << endl;
  infile.close();

  int np=npartall[0];
  int nstar=npartall[4];
  cout << "Allocating space for " << np << " gas particles and " << nstar 
       << " star particles ..." << endl;  
  PARTICLE *p=new PARTICLE[np+nstar];
  float4 mintmp=1e30, maxtmp=-1e30;
  int nskip,npstart=0,nstarstart=np;

  for(int f=0;f<numfiles;f++)
    {
     if(numfiles>1) filename=infilename+"."+dataToString(f);
     else           filename=infilename;
     bifstream infile(filename.c_str(),file_is_msb);
     if (!infile) {cout << "could not open input file! <" << filename << ">" << endl; exit(1);}
     gadget_read_header(infile,npart,massarr,time,redshift,npartall);
     if(!stars) npart[4]=0;
     nskip=npart[1]+npart[2]+npart[3];
     cout << "Reading " << npart[0] << " particles + " << npart[4] << " stars " << endl;
     cout << "Distributing at " << npstart << " and " << nstarstart << endl;
     gadget_find_block(infile,"POS ");
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       infile >> p[npstart+m].x >> p[npstart+m].y >> p[npstart+m].z;
     if(stars)
       {
       cout << "Skipping " << nskip << " Positions ..." << endl;
       if(nskip>0) infile.skip(nskip*4*3);
       if(npart[4]>0)
         {
         for (int m=0; m<npart[4]; ++m)
           infile >> p[m+nstarstart].x >> p[m+nstarstart].y >> p[m+nstarstart].z;
         }
       }

     gadget_find_block(infile,"HSML");
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       infile >> p[m+npstart].r;

     gadget_find_block(infile,lable);
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       {
       infile >> p[m+npstart].T;
       if(log10) p[m+npstart].T = log(p[m+npstart].T);
       mintmp=min(mintmp,p[m+npstart].T);
       maxtmp=max(maxtmp,p[m+npstart].T);
       p[m+npstart].type=0;
       }

     if(stars)
       {
       if(npart[4]>0)
         {
         for (int m=0; m<npart[4]; ++m)
	   {
             p[m+nstarstart].type=1;
             p[m+nstarstart].T=1;
             p[m+nstarstart].r=star_hsml;
	   }
         }
       }

     npstart+=npart[0];
     nstarstart+=npart[4];
    }

  if(minval!=maxval) 
    {
    mintmp=minval;
    maxtmp=maxval;
    }

  cout << "Range: " << mintmp << " (min) , " << maxtmp << " (max) " << endl;

  for (int m=0; m<np; ++m)
    p[m].T = (p[m].T-mintmp)/(maxtmp-mintmp);

  cout << "transforming" << endl;
  VECTOR campos(params.find<double>("camera_x"),
                params.find<double>("camera_y"),
                params.find<double>("camera_z"));
  VECTOR lookat(params.find<double>("lookat_x"),
                params.find<double>("lookat_y"),
                params.find<double>("lookat_z"));
  VECTOR sky(params.find<double>("sky_x",0),
                params.find<double>("sky_y",1),
                params.find<double>("sky_z",0));
  sky.Normalize();
  VECTOR zaxis = (lookat-campos).Norm();
  VECTOR xaxis = Cross (sky,zaxis).Norm();
  VECTOR yaxis = Cross (zaxis,xaxis);
  TRANSFORM trans;
  trans.Make_General_Transform
    (TRANSMAT(xaxis.x,xaxis.y,xaxis.z,
              yaxis.x,yaxis.y,yaxis.z,
              zaxis.x,zaxis.y,zaxis.z,
              0,0,0));
  trans.Invert();
  TRANSFORM trans2;
  trans2.Make_Translation_Transform(-campos);
  trans2.Add_Transform(trans);
  trans=trans2;
  //  trans.Add_Transform(trans2);
  double fov = params.find<double>("fov",45); //in degrees
  double fovfct = tan(fov*0.5*Pi/180.);

  for (int m=0; m<np+nstar; ++m)
    {
    VECTOR v(p[m].x,p[m].y,p[m].z);
    v=trans.TransPoint(v);
    p[m].x=v.x; p[m].y=v.y; p[m].z=v.z;
    }
  for (int m=0; m<np+nstar; ++m)
    {
    float8 xfac=1./(fovfct*p[m].z);
    p[m].x = res*.5 * (p[m].x+fovfct*p[m].z)*xfac;
    p[m].y = res*.5 * (p[m].y+fovfct*p[m].z)*xfac;
    p[m].ro = p[m].r;
    p[m].r = p[m].r *res*.5*xfac;
    }

  cout << "sorting by z" << endl;
  sort(&p[0], &p[np+nstar], zcmp());

  cout << "plotting" << endl;

  float8 rfac=1.5;
  float8 powtmp = pow(float8(Pi/2),float8(1./3.));
  float8 sigma0=powtmp/sqrt(2*Pi);
  float8 i00=brightness/(2*sqrt(Pi)*powtmp);

  arr2<COLOUR> pic(res,res);
  COLOURMAP amap,emap;
  switch(colors)
    {
    case 1:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY( 0,.2,COLOUR(1,0,0),COLOUR(0,1,0)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.2,.5,COLOUR(0,1,0),COLOUR(0,0,1)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,.65,COLOUR(0,0,1),COLOUR(0,1,1)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.65,.8,COLOUR(0,1,1),COLOUR(1,0,1)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.8,1.,COLOUR(1,0,1),COLOUR(1,1,1)));
      emap=amap;
    break;
    case 2:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.33,COLOUR(0.00,0.00,1.00),COLOUR(1.00,0.66,0.66)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,COLOUR(1.00,0.66,0.66),COLOUR(0.50,1.00,0.50)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,COLOUR(0.50,1.00,0.50),COLOUR(0.00,0.66,0.33)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,1.0,COLOUR(0.00,0.66,0.33),COLOUR(1.00,0.00,0.00)));
      emap=amap;
    case 3:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.0,.5,COLOUR(0.0,0.0,1.0),COLOUR(0.5,1.0,0.5)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,1.,COLOUR(0.5,1.0,0.5),COLOUR(1.0,0.0,0.0)));
      emap=amap;
    case 4:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.16,COLOUR(0.00,0.00,0.00),COLOUR(0.00,0.00,1.00)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.16,.33,COLOUR(0.00,0.00,1.00),COLOUR(1.00,0.66,1.00)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,COLOUR(1.00,0.66,1.00),COLOUR(0.50,1.00,1.00)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,COLOUR(0.50,1.00,1.00),COLOUR(0.00,0.66,0.66)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,1.0,COLOUR(0.00,0.66,0.66),COLOUR(1.00,0.00,0.00)));
      emap=amap;
    break;
    }

  bool gas = params.find<bool>("gas",true);

  for (int m=0; m<np+nstar; ++m)
    {
    if ((m%100000)==0)
      cout << m << " of " << np+nstar << " Particles ..." << endl;

    if (p[m].z<=0) continue;
    if (!gas and p[m].type==0) continue;
    if (!stars and p[m].type==1) continue;

    float8 temp=p[m].T;
    temp=max(float8(0.),min(float8(1.),temp));
    COLOUR a,e;
    if(p[m].type==0)
      {
      a=amap.Get_Colour(temp);
      e=emap.Get_Colour(temp);
      }
    else
      {
      a=COLOUR(1,1,0);
      e=COLOUR(1,1,0);
      }

    COLOUR q (e.r/(a.r+0.01),e.g/(a.g+0.01),e.b/(a.b+0.01));

    float8 r=p[m].r, ro=p[m].ro;
    float8 sigma=ro*sigma0;
    float8 fac0=-0.5/(sigma*sigma);
    float8 i0=i00/ro;
    float8 radsq = rfac*rfac*r*r;
    float8 prefac1 = ro/r*ro/r*fac0;
    float8 prefac2 = -0.5*i0;

    float8 posx=p[m].x, posy=p[m].y;
    int minx = max(int(posx-rfac*r),0), maxx = min(int(posx+rfac*r+1),res),
        miny = max(int(posy-rfac*r),0), maxy = min(int(posy+rfac*r+1),res);
    for (int y=miny; y<maxy; ++y)
      {
      float8 ysq=(y-posy)*(y-posy);
      for (int x=minx; x<maxx; ++x)
        {
        float8 dsq = (x-posx)*(x-posx) + ysq;
        if (dsq<radsq)
          {
          float8 fac = prefac2*xexp(prefac1*dsq);
          pic[x][y].r = q.r+(pic[x][y].r-q.r)*xexp(fac*a.r);
          pic[x][y].g = q.g+(pic[x][y].g-q.g)*xexp(fac*a.g);
          pic[x][y].b = q.b+(pic[x][y].b-q.b)*xexp(fac*a.b);
          }
        }
      }
    }

  cout << "writing" << endl;
  const byte header[18] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    res%256, res/256, res%256, res/256, 24, 32 };

  bofstream file(outfile.c_str(),file_is_natural);
  for (int m=0; m<18; ++m) file << header[m];
  for (int y=0; y<res; ++y)
    {
    for (int x=0; x<res; ++x)
      {
      pic[x][y].b=min(float8(1.), max(float8(0.), float8(pic[x][y].b)));
      byte pix = min(byte(255),byte(256*pic[x][y].b));
      file << pix;
      pic[x][y].g=min(float8(1.), max(float8(0.), float8(pic[x][y].g)));
      pix = min(byte(255),byte(256*pic[x][y].g));
      file << pix;
      pic[x][y].r=min(float8(1.), max(float8(0.), float8(pic[x][y].r)));
      pix = min(byte(255),byte(256*pic[x][y].r));
      file << pix;
      }
    }
  }
